// Import necessary functions from the D3 library.
import {
  scaleLinear, // Function for creating linear scales.
  extent, // Function to find the extent (min and max values) of a dataset.
  axisLeft, // Function for creating the y-axis.
  axisBottom, // Function for creating the x-axis.
  symbols, // Array of predefined symbol types.
  symbol, // Function for generating symbols.
  scaleOrdinal, // Function for creating ordinal scales.
} from 'd3';

// Define the scatter plot function using an arrow function.
export const scatterPlot = () => {
  // Initialize variables using the 'let' keyword.

  // Width of the SVG element.
  let width;

  // Height of the SVG element.
  let height;

  // Dataset to be visualized.
  let data;

  // Function to extract x-axis values from the data.
  let xValue;

  // Function to extract y-axis values from the data.
  let yValue;

  // Function to extract symbol values from the data.
  let symbolValue;

  // Function to extract radius x-values from the data.
  let rxValue;

  // Function to extract radius y-values from the data.
  let ryValue;

  // Function to extract animal type values from the data.
  let animal_typeValue;

  // Object defining the margins of the plot.
  let margin;

  // Radius of the data points.
  let radius;

  // Define the main function that generates the scatter plot.
  const my = (selection) => {
    // Create a linear scale for the x-axis using the scaleLinear function.
    const x = scaleLinear()
      .domain(extent(data, xValue)) // Set the domain of the scale using the extent of the x-axis values.
      .range([margin.left, width - margin.right]); // Set the range of the scale.

    // Create a linear scale for the y-axis using the scaleLinear function.
    const y = scaleLinear()
      .domain(extent(data, yValue)) // Set the domain of the scale using the extent of the y-axis values.
      .range([height - margin.bottom, margin.top]); // Set the range of the scale.

    // Create an ordinal scale for symbol values using the scaleOrdinal function.
    const symbolScale = scaleOrdinal()
      .domain(data.map(symbolValue)) // Set the domain of the scale to unique symbol values in the data.
      .range(symbols); // Set the range of the scale to predefined symbol types.

    // Define a function to assign colors to symbols based on animal type.
    const marker = (d) => {
      // Switch statement to assign colors based on animal type.
      switch (d) {
        case 'Lion (Panthera leo)':
          return '#C71585'; // Return a color for lions.
        case 'Tiger (Panthera tigris)':
          return 'range'; // Return a color for tigers.
        default:
          return 'green'; // Return a default color.
      }
    };

    // Create a symbol generator using the symbol function.
    const symbolGenerator = symbol();

    // Create an array of marks (data points) to be plotted.
    const marks = data.map((d) => ({
      // Calculate the x-coordinate of the data point.
      x: x(xValue(d)),
      // Calculate the y-coordinate of the data point.
      y: y(yValue(d)),
      // Extract the animal type value from the data.
      animal_type: animal_typeValue(d),
      // Assign a color to the data point based on animal type.
      color: marker(animal_typeValue(d)),
      // Extract the radius x-value from the data.
      rx: rxValue(d),
      // Extract the radius y-value from the data.
      ry: ryValue(d),
      // Generate the symbol path for the data point.
      pathD: symbolGenerator.type(
        symbolScale(symbolValue(d)),
      )(),
    }));

    // Select all path elements and bind the data.
    const circles = selection
      .selectAll('path')
      .data(marks)
      .join('path')
      .attr('cx', (d) => d.x) // Set the x-coordinate attribute.
      .attr('cy', (d) => d.y) // Set the y-coordinate attribute.
      .attr('fill', (d) => d.color) // Set the fill color attribute.
      .attr('rx', (d) => d.rx) // Set the rx attribute.
      .attr('ry', (d) => d.ry) // Set the ry attribute.
      .attr('r', radius) // Set the radius attribute.
      .attr('d', (d) => d.pathD) // Set the pathD attribute.
      .attr('transform', (d) => `translate(${d.x},${d.y})`); // Set the transform attribute.

    // Select all elements with class 'y-axis' and create the y-axis.
    selection
      .selectAll('.y-axis')
      .data([null])
      .join('g')
      .attr('class', 'y-axis')
      .attr('transform', `translate(${margin.left},0)`)
      .call(axisLeft(y));

    // Select all elements with class 'x-axis' and create the x-axis.
    selection
      .selectAll('.x-axis')
      .data([null])
      .join('g')
      .attr('class', 'x-axis')
      .attr(
        'transform',
        `translate(0,${height - margin.bottom})`,
      )
      .call(axisBottom(x));
  };

  // Function to set or retrieve the width of the plot.
  my.width = function (_) {
    return arguments.length ? ((width = +_), my) : width;
  };

  // Function to set or retrieve the height of the plot.
  my.height = function (_) {
    return arguments.length ? ((height = +_), my) : height;
  };

  // Function to set or retrieve the dataset to be visualized.
  my.data = function (_) {
    return arguments.length ? ((data = _), my) : data;
  };

  // Function to set or retrieve the xValue accessor function.
  my.xValue = function (_) {
    return arguments.length ? ((xValue = _), my) : xValue;
  };

  // Function to set or retrieve the yValue accessor function.
  my.yValue = function (_) {
    return arguments.length ? ((yValue = _), my) : yValue;
  };

  // Function to set or retrieve the rxValue accessor function.
  my.rxValue = function (_) {
    return arguments.length ? ((rxValue = _), my) : rxValue;
  };

  // Function to set or retrieve the ryValue accessor function.
  my.ryValue = function (_) {
    return arguments.length ? ((ryValue = _), my) : ryValue;
  };

  // Function to set or retrieve the animal_typeValue accessor function.
  my.animal_typeValue = function (_) {
    return arguments.length
      ? ((animal_typeValue = _), my)
      : animal_typeValue;
  };

  // Function to set or retrieve the symbolValue accessor function.
  my.symbolValue = function (_) {
    return arguments.length
      ? ((symbolValue = _), my)
      : symbolValue;
  };

  // Function to set or retrieve the margin object.
  my.margin = function (_) {
    return arguments.length ? ((margin = _), my) : margin;
  };

  // Function to set or retrieve the radius of data points.
  my.radius = function (_) {
    return arguments.length ? ((radius = +_), my) : radius;
  };

  // Return the main function.
  return my;
};
